"""Model254 module."""

from dataclasses import dataclass, field

import numpy as np
import pandas as pd
from numpy.typing import NDArray

from ..hlm import ModelForcings, ModelParameters, ModelStates, NodeParameters
from ..hlm_bmi import BmiFields, HlmBmi

DAYS_IN_MONTH = 30
SECONDS_IN_HOUR = 3600.0
SECONDS_IN_DAY = 86400.0
MILLIMETER_IN_METER = 1000.0

S_R = 1.0  # m. Ref val


@dataclass
class Model254NodeParameters(NodeParameters):
    """Model254 node parameters."""

    a_i: float
    l_i: float
    a_h: float
    invtau: float
    k_2: float
    k_i: float


@dataclass
class Model254States(ModelStates):
    """Model254 states."""

    q: float = 1.0e-6  # Channel discharge [m**3 s**-1]
    s_p: float = 0.0  # Water ponded on hillslope surface [m]
    s_t: float = 0.0  # Effective water depth in the top soil layer [m]
    s_s: float = 0.0  # Effective water depth in hillslope subsurface [m]
    s_precip: float = 0.0  # Total fallen precipitation from time 0 to t [m]
    v_r: float = 0.0  # Total flux of water from runoff from time 0 to t [m**3 s**-1]
    q_b: float = 1.0e-6  # Channel discharge from baseflow [m**3 s**-1]

    @classmethod
    def init(cls, arr: NDArray[np.float64]):
        """Init an instance from an array."""
        return cls(
            q=arr[0],
            s_p=arr[1],
            s_t=arr[2],
            s_s=arr[3],
            s_precip=arr[4],
            v_r=arr[5],
            q_b=max(0.001, arr[6]),
            # https://github.com/rmolina/asynch/blob/4d3e4bf281a184279147b408b1df24f2298df606/src/models/equations.c#L1814
        )

    def from_array(self, arr: NDArray[np.float64]):
        """Update an instance from an array."""
        self.q = arr[0]
        self.s_p = arr[1]
        self.s_t = arr[2]
        self.s_s = arr[3]
        self.s_precip = arr[4]
        self.v_r = arr[5]
        self.q_b = arr[6]

    def to_array(self) -> NDArray[np.float64]:
        """Convert the instance to an array."""
        return np.array(
            [self.q, self.s_p, self.s_t, self.s_s, self.s_precip, self.v_r, self.q_b],
            dtype=np.float64,
        )


@dataclass
class Model254Parameters(ModelParameters):  # pylint:disable=too-many-instance-attributes
    """Model254 global parameters."""

    v_0: float = 0.33  # m s-1
    lambda_1: float = 0.2
    lambda_2: float = -0.1
    v_h: float = 0.02  # m s-1
    k_3: float = 2.0425e-6 / 60.0  # s-1
    k_i_factor: float = 0.02
    h_b: float = 0.5  # m
    s_l: float = 0.1  # m
    a: float = 0.0
    b: float = 99.0
    exponent: float = 3.0
    v_b: float = 0.75  # m s-1


@dataclass
class Model254Forcings(ModelForcings):
    """Model254 forcings."""

    precipitation: pd.DataFrame = field(default_factory=pd.DataFrame)
    potential_evaporation: pd.DataFrame = field(default_factory=pd.DataFrame)


@dataclass
class EvaporationFluxes:
    """Evaporation fluxes dataclass."""

    e_p: float
    e_t: float
    e_s: float


@dataclass
class StorageFluxes:
    """Storage fluxes dataclass."""

    q_pl: float
    q_pt: float
    q_ts: float
    q_sl: float


@dataclass
class Model254(HlmBmi):
    """Model254 class."""

    parameters: Model254Parameters = field(default_factory=Model254Parameters)
    forcings_data: Model254Forcings = field(default_factory=Model254Forcings)
    initial_values: Model254States = field(default_factory=Model254States)

    bmi_fields: BmiFields = field(
        default_factory=lambda: BmiFields(
            component_name="HLM#254",
            input_var_names=(
                "atmosphere_water__precipitation_leq-volume_flux",
                "land_surface_water__potential_evaporation_volume_flux",
            ),
            input_var_units=(
                "m s-1",
                "m s-1",
            ),
            output_var_names=(
                "channel_discharge",
                "water_ponded_on_hillslope_surface",
                "effective_water_depth_in_the_top_soil_layer",
                "effective_water_depth_in_hillslope_subsurface",
                "total_precipitation_from_time_0_to_t",
                "total_runoff_from_time_0_to_t",
                "channel_discharge_from_baseflow",
            ),
            output_var_units=(
                "m3 s-1",
                "m",
                "m",
                "m",
                "m",
                "m",
                "m3 s-1",
            ),
        )
    )

    def equations(
        self,
        t: float,
        y: NDArray[np.float64],
        node_id: int,
        # dense_output: bool,
    ) -> NDArray[np.float64]:
        """Compute the updated derivatives using dense solutions."""

        node_param: Model254NodeParameters = self.network.nodes[node_id]["parameters"]
        states = Model254States.init(y)
        children_solutions = self.get_children_solutions(node_id, t)

        precipitation = self.get_input(self.forcings_data.precipitation, node_id, t)
        evaporation_fluxes = self.get_evaporation_fluxes(t, node_id, states)
        storage_fluxes = self.get_storage_fluxes(states, node_param)

        discharge = self.get_discharge(
            states,
            node_param,
            storage_fluxes,
            children_solutions,
        )

        ponded = (
            precipitation
            - storage_fluxes.q_pl
            - storage_fluxes.q_pt
            - evaporation_fluxes.e_p
        )
        topsoil = storage_fluxes.q_pt - storage_fluxes.q_ts - evaporation_fluxes.e_t
        subsurface = storage_fluxes.q_ts - storage_fluxes.q_sl - evaporation_fluxes.e_s

        baseflow = self.get_baseflow(
            node_param, states, children_solutions, storage_fluxes
        )

        return np.array(
            [
                discharge,
                ponded,
                topsoil,
                subsurface,
                precipitation,
                storage_fluxes.q_pl,
                baseflow,
            ]
        )

    def get_baseflow(self, node_param, states, children_solutions, storage_fluxes):
        """Get baseflow."""
        model_param = self.parameters

        base_flow = storage_fluxes.q_sl * node_param.a_h - states.q_b  # * 60.0

        for child_states, _ in children_solutions:
            child_states = Model254States.init(child_states)
            baseflow_child = child_states.q_b

            if baseflow_child < 0:
                baseflow_child = 1e-6
            base_flow += baseflow_child

        base_flow *= model_param.v_b / node_param.l_i
        return base_flow

    def get_evaporation_fluxes(
        self,
        t: float,
        node_id: int,
        states: Model254States,
    ) -> EvaporationFluxes:
        """Get evaporation fluxes."""

        model_param = self.parameters

        potential_evaporation = self.get_input(
            self.forcings_data.potential_evaporation, node_id, t
        )

        if potential_evaporation <= 0.0:
            return EvaporationFluxes(0.0, 0.0, 0.0)

        corr = (
            states.s_p
            + states.s_t / model_param.s_l
            + states.s_s / (model_param.h_b - model_param.s_l)
        )
        if corr <= 1e-12:
            return EvaporationFluxes(0.0, 0.0, 0.0)

        e_p = (states.s_p / S_R) * potential_evaporation / corr
        e_t = states.s_t / model_param.s_l * potential_evaporation / corr
        e_s = (
            states.s_s
            / (model_param.h_b - model_param.s_l)
            * potential_evaporation
            / corr
        )
        return EvaporationFluxes(e_p, e_t, e_s)

    def get_storage_fluxes(
        self,
        states: Model254States,
        node_param: Model254NodeParameters,
    ) -> StorageFluxes:
        """Get fluxes between storages."""

        model_param = self.parameters

        pow_term = (
            np.power(1.0 - states.s_t / model_param.s_l, model_param.exponent)
            if (1.0 - states.s_t / model_param.s_l > 0.0)
            else 0.0
        )
        k_t = node_param.k_2 * (model_param.a + model_param.b * pow_term)

        # Fluxes
        q_pl = node_param.k_2 * states.s_p  # q_{pc} = k_2 * s_p # [m/s]
        q_pt = k_t * states.s_p  # q_{pt} = k_t * s_p # [m/s]
        q_ts = node_param.k_i * states.s_t  # q_{ts} = k_i * s_t # [m/s]
        q_sl = model_param.k_3 * states.s_s  # q_{sc} = k_3 * s_s # [m/s]

        return StorageFluxes(q_pl, q_pt, q_ts, q_sl)

    def get_discharge(
        self,
        states: Model254States,
        node_param: Model254NodeParameters,
        storage_fluxes: StorageFluxes,
        children_solutions: list[tuple[NDArray[np.float64], float]],
    ) -> float:
        """Get discharge."""

        model_param = self.parameters
        # Discharge
        discharge = (
            -states.q  # m3/s
            + node_param.a_h
            * (storage_fluxes.q_pl + storage_fluxes.q_sl)  # m2*(m/s) == m3/s
        )

        # children_solutions[child_id] = (child_ode_result, child_flow_partition)

        for child_states, child_flow_partition in children_solutions:
            child_states_arr = Model254States.init(child_states)
            discharge_child = child_states_arr.q * child_flow_partition

            if discharge_child < 0:
                discharge_child = 1e-6
            discharge += discharge_child

        if states.q < 0 and model_param.lambda_1 < 1:
            return 0.0

        return node_param.invtau * np.power(states.q, model_param.lambda_1) * discharge

    def _calculate_additional_network_parameters(self) -> None:
        for node_id in self.network.nodes:
            model_parameters = self.parameters
            node_parameters = self.network.nodes[node_id]["parameters"]

            a_i = node_parameters[0]
            l_i = node_parameters[1] * 1e3  # km -> m
            a_h = node_parameters[2] * 1e6  # km2 -> m2
            # models/definitions/ConvertParams

            # Compute extra local parameters
            invtau = (
                model_parameters.v_0
                * pow(a_i, model_parameters.lambda_2)
                / ((1.0 - model_parameters.lambda_1) * l_i)
            )  # [1/s]

            k_2 = model_parameters.v_h * l_i / a_h  # [1/s]

            k_i = k_2 * model_parameters.k_i_factor  # [1/s]

            self.network.nodes[node_id]["parameters"] = Model254NodeParameters(
                a_i, l_i, a_h, invtau, k_2, k_i
            )

    def _load_forcing_series(self) -> None:
        super()._load_forcing_series()
        # Convert from mm/hr to m/s
        self.forcings_data.precipitation /= SECONDS_IN_HOUR * MILLIMETER_IN_METER
        # Convert from mm/month to m/s
        self.forcings_data.potential_evaporation /= (
            DAYS_IN_MONTH * SECONDS_IN_DAY * MILLIMETER_IN_METER
        )



"""Adds BMI functions to the base Hlm class."""

from dataclasses import dataclass, field

import numpy as np
from bmipy import Bmi
from numpy.typing import NDArray

from .hlm import Hlm


@dataclass
class BmiFields:
    """Required fields for every BMI-compliant model."""

    component_name: str = field(default_factory=str)
    input_var_names: tuple[str, ...] = field(default_factory=tuple)
    input_var_units: tuple[str, ...] = field(default_factory=tuple)
    output_var_names: tuple[str, ...] = field(default_factory=tuple)
    output_var_units: tuple[str, ...] = field(default_factory=tuple)


@dataclass
class HlmBmi(Hlm, Bmi):  # pylint: disable=too-many-public-methods
    """Extend the Hlm base class to make it BMI-compliant."""

    bmi_fields: BmiFields = field(default_factory=BmiFields)

    # Model control functions

    def initialize(self, config_file: str) -> None:  # DONE
        """Perform startup tasks for the model."""
        self.startup(config_file)

    def update(self) -> None:  # DONE
        """Advance model state by one time step."""
        self._run_until(self.current_time + self.config["time_step"])

    def update_until(self, time: float) -> None:  # DONE
        """Advance model state until the given time."""
        self._run_until(time)

    def finalize(self) -> None:
        """Perform tear-down tasks for the model."""
        # TODO

    # Model information functions

    def get_component_name(self) -> str:  # DONE
        """Name of the component."""
        return self.bmi_fields.component_name

    def get_input_item_count(self) -> int:  # DONE
        """Count of a model's input variables."""
        return len(self.bmi_fields.input_var_names)

    def get_output_item_count(self) -> int:  # DONE
        """Count of a model's output variables."""
        return len(self.bmi_fields.output_var_names)

    def get_input_var_names(self) -> tuple[str, ...]:  # type: ignore
        # https://github.com/csdms/bmi-python/blob/b0ef8b640aee4b499c5d7a8b3b30d625c28a2194/src/bmipy/bmi.py#L110
        """Get the model's input variables."""
        return self.bmi_fields.input_var_names

    def get_output_var_names(self) -> tuple[str, ...]:  # type: ignore
        # https://github.com/csdms/bmi-python/blob/b0ef8b640aee4b499c5d7a8b3b30d625c28a2194/src/bmipy/bmi.py#L133
        """Get the model's output variables."""
        return self.bmi_fields.output_var_names

    # Variable information functions

    def get_var_grid(self, name: str) -> int:  # DONE
        """Get grid identifier for the given variable."""
        var_names = self.get_input_var_names() + self.get_output_var_names()
        for var_idx, var_name in enumerate(var_names):
            if var_name == name:
                return var_idx
        raise KeyError

    def get_var_type(self, name: str) -> str:  # DONE
        """Get data type of the given variable."""
        return str(self.get_value_ptr(name).dtype)

    def get_var_units(self, name: str) -> str:  # DONE
        """Get units of the given variable."""
        return (self.bmi_fields.input_var_units + self.bmi_fields.output_var_units)[
            self.get_var_grid(name)
        ]

    def get_var_itemsize(self, name: str) -> int:  # DONE
        """Get memory use for each array element in bytes."""
        return np.dtype(self.get_var_type(name)).itemsize

    def get_var_nbytes(self, name: str) -> int:  # DONE
        """Get size, in bytes, of the given variable."""
        return self.get_value_ptr(name).nbytes

    def get_var_location(self, name: str) -> str:
        """Get the grid element type that the a given variable is defined on."""
        return "node"  # FIXME: *node*: A point that has a coordinate pair or triplet

    # Time functions

    def get_current_time(self) -> float:
        """Return the current time of the model."""
        return self.current_time

    def get_start_time(self) -> float:
        """Start time of the model."""
        return self.config["start_time"]

    def get_end_time(self) -> float:
        """End time of the model."""
        return self.config["end_time"]

    def get_time_units(self) -> str:
        """Time units of the model."""
        return "s"

    def get_time_step(self) -> float:
        """Return the current time step of the model."""
        return self.config["time_step"]

    # Variable getter and setter functions

    def get_value(
        self, name: str, dest: NDArray[np.float64]
    ) -> NDArray[np.float64]:  # DONE
        """Get a copy of values of the given variable."""
        dest[:] = self.get_value_ptr(name).flatten()
        return dest

    def get_value_ptr(self, name: str) -> NDArray[np.float64]:  # DONE
        """Get a reference to values of the given variable."""

        for idx, var_name in enumerate(self.get_input_var_names()):
            if var_name == name:
                return self.current_forcings[:, idx]

        for idx, var_name in enumerate(self.get_output_var_names()):
            if var_name == name:
                return self.current_states[:, idx]

        raise KeyError

    def get_value_at_indices(
        self, name: str, dest: NDArray[np.float64], inds: NDArray[np.int_]
    ) -> NDArray[np.float64]:  # DONE
        """Get values at particular indices."""
        dest[:] = self.get_value_ptr(name)[inds]
        return dest

    def set_value(self, name: str, src: NDArray[np.float64]) -> None:  # DONE
        """Specify a new value for a model variable."""

        for idx, var_name in enumerate(self.get_input_var_names()):
            if var_name == name:
                self.current_forcings[:, idx] = src[:]
                return

        for idx, var_name in enumerate(self.get_output_var_names()):
            if var_name == name:
                self.current_states[:, idx] = src[:]
                return

        raise KeyError

    def set_value_at_indices(
        self, name: str, inds: NDArray[np.int_], src: NDArray[np.float64]
    ) -> None:
        """Specify a new value for a model variable at particular indices."""
        self.get_value_ptr(name)[inds] = src

    # # Grid information

    def get_grid_rank(self, grid: int) -> int:
        """Get number of dimensions of the computational grid."""
        raise NotImplementedError("`get_grid_rank` is not implemented")

    def get_grid_size(self, grid: int) -> int:
        """Get the total number of elements in the computational grid."""
        raise NotImplementedError("`get_grid_size` is not implemented")

    def get_grid_type(self, grid: int) -> str:
        """Get the grid type as a string."""
        return "unstructured"

    # Uniform rectilinear

    def get_grid_shape(self, grid: int, shape: NDArray[np.int_]) -> NDArray[np.int_]:
        """Get dimensions of the computational grid."""
        raise NotImplementedError("`get_grid_shape` is not implemented")

    def get_grid_spacing(
        self, grid: int, spacing: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        """Get distance between nodes of the computational grid."""
        raise NotImplementedError("`get_grid_spacing` is not implemented")

    def get_grid_origin(
        self, grid: int, origin: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        """Get coordinates for the lower-left corner of the computational grid."""
        raise NotImplementedError("`get_grid_origin` is not implemented")

    # Non-uniform rectilinear, curvilinear

    def get_grid_x(self, grid: int, x: NDArray[np.float64]) -> NDArray[np.float64]:
        """Get coordinates of grid nodes in the x direction."""
        raise NotImplementedError("`get_grid_x` is not implemented")

    def get_grid_y(self, grid: int, y: NDArray[np.float64]) -> NDArray[np.float64]:
        """Get coordinates of grid nodes in the y direction."""
        raise NotImplementedError("`get_grid_y` is not implemented")

    def get_grid_z(self, grid: int, z: NDArray[np.float64]) -> NDArray[np.float64]:
        """Get coordinates of grid nodes in the z direction."""
        raise NotImplementedError("`get_grid_z` is not implemented")

    def get_grid_node_count(self, grid: int) -> int:
        """Get the number of nodes in the grid."""
        return self.network.number_of_nodes()

    def get_grid_edge_count(self, grid: int) -> int:
        """Get the number of edges in the grid."""
        return self.network.number_of_edges()

    def get_grid_face_count(self, grid: int) -> int:
        """Get the number of faces in the grid."""
        raise NotImplementedError("`get_grid_face_count` is not implemented")

    def get_grid_edge_nodes(
        self, grid: int, edge_nodes: NDArray[np.int64]
    ) -> NDArray[np.int64]:
        """Get the edge-node connectivity."""
        edge_nodes[:] = np.array(
            [node for edge in self.network.edges for node in reversed(edge)],
        )
        return edge_nodes

    def get_grid_face_edges(
        self, grid: int, face_edges: NDArray[np.int_]
    ) -> NDArray[np.int_]:
        """Get the face-edge connectivity."""
        raise NotImplementedError("`get_grid_face_edges` is not implemented")

    def get_grid_face_nodes(
        self, grid: int, face_nodes: NDArray[np.int_]
    ) -> NDArray[np.int_]:
        """Get the face-node connectivity."""
        raise NotImplementedError("`get_grid_face_nodes` is not implemented")

    def get_grid_nodes_per_face(
        self, grid: int, nodes_per_face: NDArray[np.int_]
    ) -> NDArray[np.int_]:
        """Get the number of nodes for each face."""
        raise NotImplementedError("`get_grid_nodes_per_face` is not implemented")
